#!/bin/bash


# HOP v.1.0
# Autor. F. Javier Campos
# Junio 2009

clear
echo "########################################"
echo "#                                      #"
echo "#              HOP  v.1.0              #"
echo "#                                      #"
echo -e "########################################\n"


if [ $# -lt 3 ]
then
	echo "USO DEL SCRIPT:"
	echo "    $0 < interfaz > -manual nombre_fichero_pseudonimos"
	echo "  ó bien"
	echo -e "    $0 < interfaz > -auto numero_pseudos\n\n"
	echo "Nota: <interfaz> se debe especificar también en /etc/olsrd.conf y /usr/local/etc/hip/hip.conf"
	echo "Nota 2: en modo '-auto' se genera el fichero 'pseudos.temp' con los pseudonimos elegidos."
	exit 1
fi

if [ $UID != 0 ]
then
	echo -e "ERROR.\tDebes ejecutar este script como ROOT."
	echo -e "\tPrueba a escribir:"
	echo -e "\t\tsudo $0 $1 $2 $3\n"
	exit 1
fi

# si no existen los scripts... error
if [ ! -f ./scripts/pseudonym-manager.sh ]
then
	echo "Error. El fichero 'pseudonym-manager.sh' no se encuentra en el directorio ./scripts."
	exit 1
fi
if [ ! -f ./scripts/filtrar_arp.sh ]
then
	echo "Error. El fichero 'filtrar_arp.sh' no se encuentra en el directorio ./scripts."
	exit 1
fi


# prerrequisito, tener instalado arptables
if [ ! -f /sbin/arptables ]
then
	echo "Error. Se necesita 'arptables'"
	echo "Prueba a ejecutar: 'apt-get install arptables'"
	exit 1
fi

# existe olrd ?
if [ ! -f /usr/sbin/olsrd ]
then
	echo "Error. Se necesita tener instalado 'olsrd'"
	exit 1
fi

# existe hip ?
if [ ! -f /usr/local/sbin/hip ]
then
	echo "Error. Se necesita tener instalado 'hip'"
	exit 1
fi



if [ "$2" ==  "-manual" ]
then
	# crear nuevas direcciones ip
	./scripts/pseudonym-manager.sh $1 add $3

	if [ -f $3 ] # si el fichero existe...
	then
		# filtrar ARPs
		./scripts/filtrar_arp.sh $3

		# ejecutar OLSR en modo -pseudo
		/usr/sbin/olsrd -pseudo $3 > salida_olsr.txt &

		# ejecutar HIP-MANET en modo -pseudo
		/usr/local/sbin/hip -v -V -pseudo $3
	else
		echo "Error. El fichero '$3' no existe."
		exit 1
	fi
elif [ "$2" ==  "-auto" ]
then
	# borrar direcciones ip antiguas
	./scripts/pseudonym-manager.sh $1 del $3

	echo -e "\nEjecutando OLSR con generacion automática de pseudonimos (background)\n"

	# OLSR con auto-pseudonimos
	/usr/sbin/olsrd -pseudo pseudos.temp -auto $3 > salida_olsr.txt &

	echo -e "\n\nEsperando la generación automática de pseudonimos (35 segundos)...\n"
	sleep 35 # espero a que OLSR escuche la red (+30 segundos)

	# añadir direcciones IP en la interfaz
	./scripts/pseudonym-manager.sh $1 add pseudos.temp

	# filtrar ARPs
	./scripts/filtrar_arp.sh pseudos.temp

	# ejecutar HIP-MANET en modo -pseudo
	/usr/local/sbin/hip -v -V -pseudo pseudos.temp
fi
